
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import math, warnings
from types import *
import numarray, IM
from numarray.numerictypes import *

def TestArray(shape):
    """Produces a test array

    of the following type:
        0  1  2
       10 11 12
       20 21 22
    If len(shape) == n, then the numbers have n digits. The numbers in shape
    should be < 10.
    """
    rank = len(shape)
    if rank == 0:
        return numarray.array([])
    if rank == 1:
        return numarrange.arange(shape[0])
    a = numarray.arange(shape[0])
    b = numarray.arange(shape[1])
    arr = numarray.add.outer(10*a, b)
    for i in range(2, rank):
        c = numarray.arange(shape[i])
        arr = numarray.add.outer(10*arr, c)
    return arr

def _adjust_blockspecs(blockspecs, arrayshape):
    """Convert "blockspecs" to a list of triples only.

    "blockspecs" is a list of items of one of the following forms:
         step
         (step,)
         (start, step)
         (start, step, stop)

    where the variables are all ints >= 0. The tuples may also be lists. 
    These are interpreted like range(start, stop, step). But note that the
    order is different. The blocks start with start, start+step, etc. For
    example, if the indexes are (0, 1, 2, 3, 4, 5, 6, 7, 8, 9) and the
    triple is (1, 8, 3), then range(1, 8, 3) is [1, 4, 7] and the sets of
    indexes in each interval are (1, 2, 3) and (4, 5, 6). In most cases,
    stop will be a dimension of the array.

    For each i:
        step or (step,) is interpreted as (0, step, arrayshape[i])
        (start, step) is interpreted as (start, step, arrayshape[i])
    """
    if not isinstance(blockspecs, (list, tuple)):
        raise TypeError('"blockspecs" must be a list or tuple')
    if not isinstance(arrayshape, (list, tuple)):
        raise TypeError('"arrayshape" must be a list or tuple')
    if len(blockspecs) != len(arrayshape):
        raise TypeError('"blockspecs" and "arrayshape" must have the same length')

    bs = []
    for i, blockspec in enumerate(blockspecs):
        if isinstance(blockspec, int):
            bs.append( (0, blockspec, arrayshape[i]) )
        elif isinstance(blockspec, (list, tuple)):
            if len(blockspec) == 1:
                bs.append( (0, blockspec[0], arrayshape[i]) )
            elif len(blockspec) == 2:
                bs.append( (blockspec[0], blockspec[1], arrayshape[i]) )
            elif len(blockspec) == 3:
                bs.append(tuple(blockspec))
            else:
                raise TypeError('length of "blockspec" greater than 3')
        else:
            raise TypeError('each blockspec must be an int, list, or tuple')
    return tuple(bs)

def _create_blockspecs(blockshape, dim):
    """Create default blockspecs"""
    if type(blockshape) is IntType:
        blockshape = dim * [blockshape]
    if len(blockshape) != dim:
        raise TypeError('"blockshape" must be an int or have length "dim"')
    bs = []
    for n in blockshape:
        bs.append( (0, n) )
    return bs

def count_array(shape, natype=UInt8):
    """Creates an array containing consecutive ints"""
    array = numarray.zeros(shape, natype)
    array.flat = numarray.arange(array.nelements()).astype(natype)
    return array

def BlockReduce(array, blockspecs, ufunc):
    """Apply ufunc.reduce to blocks in an array.

    "blockspecs" is a list of items of one of the following forms:
         step
         (step,)
         (start, step)
         (start, step, stop)

    where the variables are all ints >= 0. The tuples may also be lists.
    These are interpreted like range(start, stop, step). But note that the
    order is different. The blocks start with start, start+step, etc. For
    example, if the indexes are (0, 1, 2, 3, 4, 5, 6, 7, 8, 9) and the
    triple is (1, 8, 3), then range(1, 8, 3) is [1, 4, 7] and the sets of
    indexes in each interval are (1, 2, 3) and (4, 5, 6). In most cases,
    stop will be a dimension of the array.
    """
    dims = len(array.shape)
    if len(blockspecs) != dims:
        raise TypeError('blockspecs must be same length as shape')
    bs = _adjust_blockspecs(blockspecs, array.shape)
    subdims = []
    for i in range(len(bs)):
        start, step, stop = bs[i]
        size = array.shape[i]
        end = ((size-start)//step)*step + start
        subdims.append(slice(start, end, 1))
    subarr = array[subdims]
    for i in range(dims):
        subarr = subarr.copy()
        start, step, stop = bs[i]
        # Add a new dimension of size step.
        newshape = (subarr.shape[0] // step, step) + subarr.shape[1:]
        subarr.shape = newshape
        subarr = ufunc.reduce(subarr, 1)
        dims = len(subarr.shape)
        # (0,1,2,3) --> (1,2,3,0)
        perm =  tuple(range(1, dims)) + (0,)
        subarr = numarray.transpose(subarr, perm)
    return subarr

def MultiReduce(array, ufunc):
    """Reduce the array to a number using ufunc.reduce."""
    for n in range(len(array.shape)):
        array = ufunc.reduce(array)
    return array

def BlockMean(array, blockspecs):
    """Average over blocks in an array.

    The array should be floating point or complex.

    "blockspecs" is a list of items of one of the following forms:
         step
         (step,)
         (start, step)
         (start, step, stop)

    where the variables are all ints >= 0. The tuples may also be lists.
    These are interpreted like range(start, stop, step). But note that the
    order is different. The blocks start with start, start+step, etc. For
    example, if the indexes are (0, 1, 2, 3, 4, 5, 6, 7, 8, 9) and the
    triple is (1, 8, 3), then range(1, 8, 3) is [1, 4, 7] and the sets of
    indexes in each interval are (1, 2, 3) and (4, 5, 6). In most cases,
    stop will be a dimension of the array.
    """
    blockspecs = _adjust_blockspecs(blockspecs, array.shape)
    natype = array.type()
    array = IM.ArrayToArrayCastUpwards(array, IM.MODULE.NUMARRAY, Float64)
    array = BlockReduce(array, blockspecs, numarray.add)
    winsize = 1
    for triple in blockspecs:
        winsize *= triple[1]
    array = array / winsize
    return IM.ArrayToArrayCast(array, IM.MODULE.NUMARRAY, natype)

def NanPresent(arr):
    """Is there a NAN in an array?"""
    if str(Sum(arr)) == 'nan':
        return True
    return False

def CountNonZero(arr):
    """Count the non-zero elements in an array."""
    arrct = numarray.ravel(arr != 0).astype(Int)
    return numarray.sum(arrct)

def CountZeros(arr):
    """Count the zero elements in an array."""
    arrct = numarray.ravel(arr == 0).astype(Int)
    return numarray.sum(arrct)

def Stretch(arr, lo, hi, outlo=None, outhi=None, **err):
    """Linearly changes the elements of an array.

    The input array is changed linearly so lo --> outlo and hi --> outhi.

    This function can be called in two ways.
        Stretch(arr, lo, hi, outlo, outhi, **err)
        Stretch(arr, outlo, outhi, **err)
            In this case, "lo" and "hi" are set to the minimum and the
            maximum of the input array.

    "outlo" can also have the value 'min' and "outhi" can have the value
    'max' where 'min' and 'max' are strings.

    There is a keyword argument "err=". It is used to decide if an array
    varies so slightly that it can be considered to be constant. For int
    types the default err is 0. For Float32 it is 1.0e-6. For Float64, it is
    1.0e-15. These values are guesses.

    If an array has constant value x, it is clipped to be between outlow
    and outhi.
    """
    if outhi is None and outlo is not None:
        raise TypeError('if either outlo or outhi is None '
                 'then both must be None')
    natype = arr.type()
    if natype in (Bool, Complex32, Complex64):
        raise TypeError('use only int or float types for input array')
    if err.has_key('err'):
        errval = err['err']
    else:
        if isinstance(natype, IntegralType):
            errval = 0
        elif natype == Float32:
            errval = 1.0e-6      # Guesses.
        elif natype == Float64:
            errval = 1.0e-15
    if outlo is None or outlo == "min" or outhi is None or outhi == "max":
        m, M = arr.min(), arr.max()
    if outlo is None or outlo == "min":
        inlo = m
        outlo = lo
    else:
        inlo = lo
    if outhi is None or outhi == "max":
        inhi = M
        outhi = hi
    else:
        inhi = hi

    if abs(inhi - inlo) <= errval:
        return numarray.clip(arr, outlo, outhi)

    # Suppose I want to stretch a real array and then clip-cast it to ints
    # from 0 to 255. I do not want the stetched array to take the exact value
    # 256.0. So I actually stretch it from 0.0 to 256.0 - DBL_EPSILON where
    # DBL_EPSILON is the smallest amount which the mantissa part of a double
    # can change by. For my system, the C macro DBL_EPSILON is exactly
    # 2**-52 or 0x1p-52 in the %a format in C99.
    DBL_EPSILON = 2.22044604925031E-16
    outhi2 = outhi*(1.0 - DBL_EPSILON)
    slope = float(outhi2-outlo) / float(inhi-inlo)
    arr2 = arr.astype(Float64)
    arr2 = slope * (arr2 - inlo) + outlo
    return IM.ClipCast(arr2, natype)

def StretchColors(arr, lo, hi, outlo=None, outhi=None, **err):
    """Does 3 Stretches then combines them.

    "lo", "hi", "outlo", and "outhi" are triples or None. The triples can
    contain numbers or None. See "Stretch", above, for further information.
    """
    if outhi is None and outlo is not None:
        raise TypeError('if either outlo or outhi is None '
              'then both must be None')
    if outlo is None:
        outlo = (None, None, None)
    if outhi is None:
        outhi = (None, None, None)
    if not isinstance(lo, (list, tuple)) or \
       not isinstance(hi, (list, tuple)) or \
       not isinstance(outlo, (list, tuple)) or \
       not isinstance(outhi, (list, tuple)):
        raise TypeError('lo, hi, outlo, and outhi must be lists, tuples or None')
    if len(lo) != 3 or len(hi) != 3 or len(outlo) != 3 or len(outhi) != 3:
       raise typeError('lo, hi, outlo, and outhi must be None or have length 3')
    out = []
    for i in range(3):
        arr2 = Stretch(arr, lo[i], hi[i], outlo[i], outhi[i], **err)
        out.append(arr2)
    return numarray.array(out)

def Stretch3pt(arr, triple_in, triple_out):
    """Stretches an array relative to three values.

    Each triple has the form (low, middle, high). For triple_in, low may be
    the string 'min' and high may be the string 'max'.

    The input array is changed in a piecewise linear manner so that
        in_low    --> out_lo
        in_middle --> out_middle
        in_high   --> out_high
    """
    in_low, in_middle, in_high = triple_in
    out_low, out_middle, out_high = triple_out
    if in_low == 'min':
        in_low = arr.min()
    if in_high == 'max':
        in_high = arr.max()
    if not in_low <= in_middle <= in_high:
        raise ValueError('input triple not in increasing order: (%s, %s, %s)' %
           (in_low, in_middle, in_high))
    if not out_low <= out_middle <= out_high:
        raise ValueError('output triple not in increasing order: (%s, %s, %s)' %
           (out_low, out_middle, out_high))
    arr1 = numarray.where(arr < in_middle, arr, in_middle)
    arr1 = Stretch(arr1, in_low, in_middle, out_low, out_middle)
    arr2 = numarray.where(arr >= in_middle, arr, in_middle)
    arr2 = Stretch(arr2, in_middle, in_high, out_middle, out_high)
    # The "<" is important. It insures that
    #    arr[r, c] == in_middle ==> output[r, c] == out_middle
    return numarray.where(arr < in_middle, arr1, arr2) 

def InversePermutation(perm):
    """Finds the inverse of a permutation

    The input may be a list, tuple, or one-dimensional array.
    """
    numar = False
    if isinstance(perm, numarray.numarraycore.NumArray):
        numar = True
        perm = perm.tolist()
    perm = list(perm)
    a = [perm, range(len(perm))]
    a = map(None, *a)
    a.sort()
    a = map(None, *a)
    if numar:
        return numarray.array(a[1])
    else:
        return a[1]

def BinnedHistogram(arr, bins):
    """Histograms array into "bins"

    "bins" is a list of numbers in increasing order. Since both endpoints
    need to be included, the length of the output is len(bins) - 1.
    A value x is put in the i'th bin if bin[i] <= x < bin[i+1].
    """
    out = []
    # lo and hi will be int32.
    lo = numarray.where(arr < bins[0], 1, 0)
    for j in range(len(bins)-1):
        hi = numarray.where(arr < bins[j+1], 1, 0)
        out.append(Sum(hi - lo))
        lo = hi
    return out

def BinLUT(arr, bins, lut):
    """Reduce the number of values in an array.

    "bins" is a list of numbers in increasing order. Since both endpoints
    need to be included in "bins", the length of "lut" must be len(bins) - 1.
    If x is the value in the array, bins[i] <= x < bins[i+1], and lut[i] = y,
    then x will be replaced by y in the array.
    """
    if len(lut) != len(bins) - 1:
        raise TypeError('len(lut) != len(bins) + 1')
    arr2 = numarray.searchsorted(bins, arr)
    arr2 = numarray.choose(arr2, lut)
    return arr2

def LookUpTable(arr, table):
    """Passes the elements of "array" through a look up table (LUT).

    "array" must have type UInt8.
    """
    if arr.type() != UInt8:
        raise TypeError('array must have type UInt8')
    table = numarray.array(table, UInt8)
    outarr = numarray.zeros(arr.shape, UInt8)
    outarr.flat = numarray.take(table, numarray.ravel(arr))
    return outarr

def Zoom(arr, n):
    """Enlarge an array by replicating its elements.

    n can be an int or a list of ints of length len(arr.shape).
    """
    if type(n) is IntType:
        n = [n] * len(arr.shape)
    elif len(n) != len(arr.shape):
        raise ValueError('second argument to Zoom is wrong length')
    for i in range(len(arr.shape)):
        arr = numarray.repeat(arr, n[i], i)
    return arr

def Zoom1(arr, n):
    """Zoom except for first axis (for color images)

    n can be an int or a list of ints of length len(arr.shape-1).
    """
    if len(arr.shape) != 3:
        return Zoom(arr, n)
    if type(n) is IntType:
        n = [n] * (len(arr.shape) - 1)
    elif len(n) != len(arr.shape) - 1:
        raise ValueError('second argument to Zoom is wrong length')
    for i in range(1, len(arr.shape)):
        arr = numarray.repeat(arr, n[i-1], i)
    return arr

def StripEdges(array, blockshape):
    """Find a maximal  subarray completely filled by blocks of the given shape"""
    dims = len(array.shape)
    if type(blockshape) is IntType:
        blockshape = dims * [blockshape]
    if len(blockshape) != dims:
        raise TypeError('blockshape must be same length as shape')

    slicings = []
    for i in range(len(array.shape)):
        dim = array.shape[i]
        blk = blockshape[i]
        rem = dim % blk
        lo = rem/2
        hi = dim - (rem - lo)
        slicings.append(slice(lo, hi))
    slicings = tuple(slicings)
    return array[slicings]

def Shrink(arr, shrinkfactor):
    """Shrink the size of an array by shrinkfactor.

    if "shrinkfactor" is a list or tuple, it gives the amount of shrinkage
    in each dimension. It it is an int, arr is shrunk by the same ratio in
    every dimension.

    Any extra elements around the edges are removed.
    """
    if not isinstance(shrinkfactor, (int, list, tuple)):
        raise TypeError('"shrinkfactor" must be an int, list or tuple')
    dims = len(arr.shape)
    if type(n) is IntType:
        shrinkfactor = [shrinkfactor] * dims
    elif len(shrinkfactor) != dims:
        raise ValueError('second argument to Shrink is wrong length')
    subarray = StripEdges(arr, shrinkfactor)
    blockspecs = _create_blockspecs(shrinkfactor, dims)
    return BlockMean(subarray, blockspecs)

# Shrink except for first axis, for color images.
def Shrink1(arr, shrinkfactor):
    """Shrink array by factor of shrinkfactor except for axis 0 (for color images)."""
    if not isinstance(shrinkfactor, (int, list, tuple)):
        raise TypeError('"shrinkfactor" must be an int, list or tuple')
    if isinstance(shrinkfactor, (list, tuple)) and \
                    len(shrinkfactor) != len(arr.shape) - 1:
        raise TypeError('len(shrinkfactor) != len(arr.shape) - 1')
    dims = len(arr.shape)
    if type(shrinkfactor) is IntType:
        blockshape = (1,) + (dims-1) * (shrinkfactor,)
    else:
        blockshape = (1,) + tuple(shrinkfactor)
    subarray = StripEdges(arr, blockshape)
    blockspecs = _create_blockspecs(blockshape, dims)
    return BlockMean(subarray, blockspecs)

def BinarySplit(arr, n):
    """If x == BinarySplit(arr, n), then there are n ones in arr <= x.

    There is a problem with ties. Suppose arr is [0 1 2 2 3 4], then
    BinarySplit(arr, 3) is 2.0 and arr <= 2 is [1 1 1 1 0 0].

    Deprecated. Use FindThresh which is the simple program that I should
    have written to start with.
    """
    if arr.type() is Bool:
        raise TypeError('type Bool not allowed')
    if isinstance(arr.type(), IntegralType):
        _EPS = .5
    else:
        _EPS = 10E-6
    def binarysplit(arr, lo, hi, n, d):
        if abs(hi-lo) < _EPS:
            return hi
        mid = (lo+hi)/2.0
        if d > 20:
            raise Exception('recursion too deep')
        b = (arr <= mid)
        midsum = Sum(b.astype(Int))
        if n >= midsum + 1:
            return binarysplit(arr, mid, hi, n, d+1)
        elif n < midsum:
            return binarysplit(arr, lo, mid, n, d+1)
        else:
            return mid

    m, M = arr.min(), arr.max()
    size = 1
    for dim in arr.shape:
        size *= dim
    if n == size:
        return M
    if n == 0:
        return m - 1
    if n < 0 or n > size:
        raise ValueError('n must be 0 <= n <= size')
    x = binarysplit(arr, m, M, n, 0)
#    if isinstance(arr.type(), IntegralType):
#        return int(math.ceil(x))
    return x

def FindThresh(arr, n):
    """If x == FindThresh(arr, n), then there are n ones in the boolean array "arr <= x".

    If there are ties, the statement above may be slightly wrong. Suppose
    arr is [0 1 1 1 2], then x = FindThresh(arr, 3) gives x == 1.
        arr <= x is [1 1 1 1 0]
        arr < x is [1 0 0 0 0]
    """
    if n < 0 or n > arr.nelements():
        raise ValueError('n must be 0 <= n <= size')
    arr2 =  numarray.sort(numarray.ravel(arr))
    return arr2[n]

def AllIndices(shape):
    """Lists all possible indices for an array of a given shape"""
    ind = numarray.indices(shape)
    sh = ind.shape
    prod = numarray.product(shape)
    ind = numarray.reshape(ind, (sh[0], prod) )
    ind = numarray.transpose(ind)
    return ind

def IndicesArray(condition, array=None):
    """Representation for sparse arrays.

    For example, if condition[i,j] is true, (array[i,j], i, j) is a column
    in the answer. The output may have a longer type than the input.
    If array is None, the columns are (i, j).
    """
    indices = numarray.indices(condition.shape)
    flatcon = numarray.ravel(condition)
    if array is None:
        planelist = []
    else:
        first = numarray.compress(flatcon, numarray.ravel(array))
        planelist = [first]
    for i in range(indices.shape[0]):
        a = indices[i].flat
        a = numarray.compress(flatcon, a)
        planelist.append(a)
    if array is None:
        typeout = flatcon.type()
    else:
        if array.type() in (Bool, Int8, Int16):
            typeout = Int32
        elif array.type() in (UInt8, UInt16):
            typeout = UInt32
        else:
            typeout = array.type()
    return numarray.array(planelist, typeout)

def ArrayWithIndices(array, extras=tuple()):
    """Add the indices of an array as a new dimension of the array.

    For example, (array[i,j], i, j) is a column in the answer. For each
    index in extras, an extra layer is added containing only that index. If
    extras == (5, 6), then (array[i,j], i, j, 5, 6) is a column in the
    answer. The output may have a longer type than the input. I have used
    the extras only once & I doubt that they will be useful again.
    """
    oldshape = array.shape
    indices = numarray.indices(oldshape)
    array.shape = (1,) + oldshape
    slices = [array, indices]
    for index in extras:
        a = numarray.zeros(array.shape) + index
        slices.append(a)
    a = numarray.concatenate(slices)
    array.shape = oldshape
    return a

def CompressWithIndices(condition, array=None):
    """Representation for sparse arrays.

    For example, if condition[i,j] is true, (array[i,j], i, j) is a row
    in the answer. The output may have a longer type than the input.
    If array is None, the columns are (i, j).
    """
    if array is not None and condition.shape != array.shape:
        raise ValueError('"condition" and "array" must be the same shape')
    arr = IndicesArray(condition, array)
    rotate = range(len(arr.shape))
    rotate = tuple(rotate[1:] + [rotate[0]])
    return numarray.transpose(arr, rotate)

def ArrayToListWithIndices(condition, array=None):
    """Representation for sparse arrays.

    For example, if condition[i,j] is true, (array[i,j], i, j) is one of the
    tuples in the answer. If array is None, the tuples are (i, j).
    """
    if array is not None and condition.shape != array.shape:
        raise ValueError('"condition" and "array" must be the same shape')
    arr = IndicesArray(condition, array)
    rotate = range(len(arr.shape))
    rotate = tuple(rotate[1:] + [rotate[0]])
    if array is None:
        arr = numarray.transpose(arr, rotate)
        return arr.astype(Int32).tolist()
    else:
        out = [None]  # Seps up map trick.
        out.append(arr[0].tolist())
        for i in range(1, arr.shape[0]):
            out.append(arr[i].astype(Int32).tolist())
        print 'L478', len(out), out[1]
        # Transpose the list.
        return apply(map, out)

def ArrayFromIndices(indices, shape):
    """Create Bool array from list of non-zero indices."""
    out = numarray.zeros(shape, Bool)
    for coords in indices:
        out[coords] = 1
    return out

def ArrayToDict(condition, array=None):
    """Representation for sparse arrays using a dictionary.

    Suppose the dictionary is "sparse". If condition[i,j] is true,
    sparse[(i,j)] = array[i,j]. If array is None, sparse[(i,j)] = None
    """
    sparse = dict()
    arr = IndicesArray(condition)
    arr = numarray.transpose(arr)
    alist = arr.tolist()
    if array is None:
        for indices in alist:
            sparse[tuple(indices)] = None
    else:
        for indices in alist:
            tup = tuple(indices)
            sparse[tup] = array[tup]
    return sparse

def ArrayFromDict(datadict, shape, natype=Int32):
    """Create array from a dictionary mapping indices to values."""
    out = numarray.zeros(shape, natype)
    for coords in datadict.keys():
        out[coords] = datadict[coords]
    return out

def PixelNotZero(arr):
    """Replace each non-zero pixel by a pixel where each color has the value one.

    Suppose a color image is represented by an array of shape (3,128,128).
    Let an input pixel be (R,G,B) and the output pixel be (R', G', B').
    If (R,G,B) == (0,0,0), (R', G', B') == (0,0,0). Otherwise,
    (R', G', B') == (1,1,1).
    """
    t = numarray.logical_or.reduce(arr != 0, 0)
    return numarray.array(arr.shape[0] * [t])

def Threshold(array, thresh):
    return numarray.where(numarray.less(array, thresh), 0, 1).astype(Bool)

# Any value < lo is replaced by "lovalue". "lovalue" defaults to "lo".
def SaturateLow(array, lo, loval=None):
    if loval is None:
        loval = lo
    if lo is not None:
        array = numarray.where(array > lo, array, loval)
    return array

# Any value > hi is replaced by "hival". "hival" defaults to "hi".
def SaturateHigh(array, hi, hival=None):
    if hival is None:
        hival = hi
    if hi is not None:
        array = numarray.where(array < hi, array, hival)
    return array

# If lo is None, There is no saturation from below (lo = -INFINITY).
# If hi is None, There is no saturation from above (hi = INFINITY).
def Saturate(array, lo=None, hi=None):
    if lo is not None:
        array = numarray.where(array > lo, array, lo)
    if hi is not None:
        array = numarray.where(array < hi, array, hi)
    return array
