#! /usr/bin/env python

"""Tries to open an image."""

"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os, sys, time, tempfile, stat, Image, imghdr, optparse, subprocess
import walk

# Uses Image.open to open an image file. "djpeg" and "giffix" are used if
# needed. The code is ugly.

#>=== class messages ==============================================

# Prints messages to user or file. "error_file_object" must be an open file
# object.
class messages:
    """Processes error messages."""
    def __init__(self, filename, error_file_object=sys.stdout):
        self.name = filename
        self.error_file_object = error_file_object
        self.printed = 0

    def output(self, message=''):
        """print the filename and message"""
        if self.printed == 0:
            self.printed = 1
            self.error_file_object.write(self.name + ':\n')
        message = str(message)
        if len(message) != 0:
            lines = message.splitlines()
            message = '    %s\n' % '\n    '.join(lines)
            self.error_file_object.write(message)

"""The line printed by jpeginfo looks like:
"%s %4d x %-4d %2dbit %-7s %c %7ld %s" % \
    (fullname, width, height, bits, code, progressive, fs, rest)
where code is in ('Exif', 'JFIF', 'Adobe', 'n/a') and progressive is in
('N', 'P'). If the image is good, rest == '[OK]'. If the image is bad, rest
is is an error message followed by '[WARNING]' or '[ERROR]'. If the image
does not start with the correct magic, the line is
"%s %s [ERROR]" % (fullname, message).
"""
def jpegcheck(name):
    command = 'jpeginfo -c %s' % name
    child_stdin, child_stdout, child_stderr = os.popen3(command)
    s = child_stderr.read().strip()
    if len(s.strip()) != 0:
        raise Exception('error in jpegcheck: %s' % s)
    s = child_stdout.read().strip()
    if s.endswith(' [OK]'):
        return 'OK', ''
    parts = s.split()
    fullname = os.path.abspath(name)
    if os.path.abspath(parts[0]) == fullname and parts[2] ==  'x' and \
              parts[6] in ('N', 'P'):
        message = ' '.join(parts[8:-1])
        if parts[-1] == '[WARNING]':
            return 'WARNING', message
        elif parts[-1] == '[ERROR]':
            return 'ERROR', message
        else:
            print >> sys.stderr, s
            raise ValueError('BUG1: message should end with "[ERROR]" or '
                 '"[WARNING]"')
    else:
        if parts[-1] != '[ERROR]':
            print >> sys.stderr, s
            raise ValueError('BUG2: this message should end with "[ERROR]"')
        message = ' '.join(parts[1:-1])
        return 'BADMAGIC', message

def identify(filename):
    """Tries to identify the format of an image.

    Uses "identify" from ImageMagick.
    """
    p = subprocess.Popen(('identify', filename),
         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    outtext = p.stdout.read().strip()
    p.stdout.close()
    errtext = p.stderr.read().strip()
    p.stderr.close()
    format = None
    message = None
    if len(outtext) > 0:
       # Format recognized.
       words = outtext.split()
       format = words[1]
    if len(errtext) > 0:
       start = len('identify: ')
       message = errtext[start:]
    return format, message

# Opens an image with a lot of checks.
# "filename" is the image to be opened.
# "error_file_object" is an open file object such as is returned by the
# built in function "open". Another example is "sys.stdout".
# The message codes are
# 'name'	Always print the file name.
# 'errors'	Print information about broken images.
# 'verbose'	Print a lot of information.
# "codes" is a list (or tuple) of message codes. A single code need not be
# put in a list.
# A tuple (im, error) is returned. If the image does not exist or cannot be
# opened, (None, None) is returned. If the image is good, error is None.
def open_image_file(filename, error_file_object, codes=['errors']):
    """Opens an image with a lot of checks."""
    if isinstance(codes, str):
        codes = [codes]
    if 'verbose' in codes and not 'errors' in codes:
        codes.append('errors')

    V = messages(filename, error_file_object)
    if 'name' in codes:
        V.output()

    # "extcode" used only for error messages.
    root, ext = os.path.splitext(filename)
    if ext.lower() == '.jpg' or ext.lower() == '.jpeg':
        extcode = 'JPEG'
    elif ext.lower() == '.gif':
        extcode = 'GIF'
    else:
        extcode = ''

    # Some easy cases.
    if filename == '':
        if 'verbose' in codes:
            V.output("Filename is ''")
        return (None, 'File name has length zero.')
    if filename.find('.xvpics') != -1:
        if 'verbose' in codes:
            V.output('.xvpics file.')
        return (None, 'XV thumbnail')
    if not os.path.exists(filename):
        if 'verbose' in codes:
            V.output('File does not exist')
        return (None, 'File does not exist')
    if not os.path.isfile(filename):
        if 'verbose' in codes:
            V.output('File is a directory or some other non-regular file')
        return (None, 'File is a directory or some other non-regular file')
    bytes = os.stat(filename)[stat.ST_SIZE]
    if bytes < 10:
        if 'verbose' in codes:
            V.output('File has less than 10 bytes')
        return (None, 'File has less than 10 bytes')

    format, errmessage = identify(filename)

    if format == 'JPEG' and errmessage is not None:
        # Try djpeg.
        try:
            tmpout, tmpname = tempfile.mkstemp(suffix='.ppm')
            err = open(os.devnull, 'w')
            rtncode = subprocess.call(('djpeg', filename), stdout=tmpout,
                       stderr=err)
            err.close()
            os.close(tmpout)
            im = Image.open(tmpname)
            if 'errors' in codes:
                V.output('Image returned: %s %s %s' % \
                      (im.format, im.mode, im.size) )
            return (im, errmessage)
        finally:
            if os.path.exists(tmpname):
                os.unlink(tmpname)

    # Other images.
    try:
        im = Image.open(filename)
        format = im.format
    except KeyboardInterrupt:
        raise
    except MemoryError:
        raise
    except SystemError:
        raise
    except IOError, x:
        # This should pick up most non-image files.
        if 'errors' in codes:
            V.output('PIL IOError from Image.open, file ignored.')
            V.output('Message was: %s' % str(x))
        return (None, 'Probably not an image file.')
    else:
        try:
            # Some errors show up only when the opened file is used.
            pix = im.getpixel( (0,0) )
            if 'verbose' in codes:
                V.output('Image returned: %s %s %s.' % \
                  (im.format, im.mode, im.size) )
            return (im, None)
        except KeyboardInterrupt:
            raise
        except MemoryError:
            raise
        except SystemError:
            raise
        except StandardError, x:
            if 'errors' in codes:
                V.output('Unusual format or non-image file.')
                V.output('PIL IOError when loading image. Message was:')
                V.output(str(x))
                V.output('imghdr.what claims format is %s' % \
                      imghdr.what(filename))
                V.output('FIX THIS IMAGE BY HAND using xv, GIMP or some similar program.')
                V.output('No image returned.')
            return (None, str(x))
    V.output('Exited at end.')

def checkims(names, codes=['errors']):
    """Checks image files for corruption."""
    if isinstance(types, str):
        names = [names]
    for name in names:
        im, error = open_image_file(name, sys.stdout, codes)
        print '==>>', str(error)
        print

def checkims_in_file(filename, codes=['errors']):
    """Checks images listed in a file for corruption."""
    lines = open(filename, 'r').readlines()
    names = []
    for line in lines:
        name = line.strip()
        if name == '':
            continue
        names.append(name)
    checkims(names, codes)

def safeopen(filename):
    """opens an image with checks for corruption."""
    im, error = open_image_file(filename, sys.stderr, ('errors',))
    if error == None:
        return im
    else:
        return None

badbasedirname = '.badims'
def fixim(filename):
    """Fixes an image and saves it.

    The corrupted image is moved to a hidden subdirectory.
    """
    im, message = open_image_file(filename, sys.stderr, [])
    if im is None:
        return 'CANNOT FIX %s\n  %s' % (filename, message)
    if message is None:
        # Image is not broken.
        return
    head, tail = os.path.split(filename)
    badimsdir = os.path.join(head, badbasedirname)
    if not os.path.isdir(badimsdir):
        os.mkdir(badimsdir)
    badname = os.path.join(badimsdir, tail)

    i = 1
    if os.path.exists(badname):
        while os.path.exists('%s.%i' % (badname, i)):
           i += 1
        badname = '%s.%i' % (badname, i)
    os.rename(filename, badname)
    im.save(filename)

def checkims_all(output_file, badfiles, topdirs, codes=['errors']):
    out = open(output_file, 'w')
    bad = open(badfiles, 'w')
    counter = 0
    errors = 0
    t = time.time()
    for name in walk.WalkImages(topdirs):
        im, error = open_image_file(name, out, codes=codes)
        if im is None or error is not None:
            lines = error.splitlines()
            print 'L293 %s\n    %s' % (name, '\n    '.join(lines))
            bad.write('%s\n' % name)
            bad.flush()
            out.flush()
        counter += 1
        if counter % 500 == 0:
            print '%i %5.3f secs. At  %s' % (counter, time.time() - t, name)
    print counter, 'images checked.'

if __name__ == '__main__':
    usage = 'USAGE: ./checkims.py [OPTIONS] <topdir> <detailed output> <output list>'
    parser = optparse.OptionParser(usage=usage)
    parser.add_option('-n', '--name', default=False,
         action='store_true',
         help='print image name even when image is OK')
    parser.add_option('-v', '--verbose', default=False,
         action='store_true',
         help='print lots of stuff')
    options, args = parser.parse_args()
    if len(args) != 3:
        parser.print_help()
        sys.exit('three arguments required')
    topdir = args[0]
    output_file = args[1]
    badfilelist = args[2]
    codes = ['errors']
    if options.name:
        codes.append('name')
    if options.verbose:
        codes.append('verbose')
    checkims_all(output_file, badfilelist, topdir, codes)
