
import os

def prepare_excludes(excluded_dirs):
    """Normalize names of excluded directories"""
    if excluded_dirs is None:
        return ()
    if isinstance(excluded_dirs, str):
        excludes = (excluded_dirs,)
    excludes = set()
    for name in excluded_dirs:
        name = os.path.abspath(name)
        if not os.path.isdir(name):
            raise IOError('a name in "excluded_dirs" is not the name of a directory: %s' % name)
        excludes.add(name)
    return excludes                         

def prepare_topdirs(topdirs):
    """Normalize names of the topdirs"""
    if isinstance(topdirs, str):
        topdirs = (topdirs,)
    topdirs2 = list()
    for topdir in topdirs:
        topdir = os.path.abspath(topdir)
        if topdir in topdirs2:
            continue
        if not os.path.exists(topdir):
            raise IOError('topdir does not exist: %s' % topdir)
        else:
            if not os.path.isdir(topdir):
                raise IOError(
                  'topdir exists but is not a directory: %s' % topdir)
        topdirs2.append(topdir)
    for topdir1 in topdirs2:
        for topdir2 in topdirs2:
            if topdir1 != topdir2 and isunder(topdir1, topdir2):
                raise TypeError('a topdir cannot be a subdirectory '
                  'of another topdir:\n  "%s"\n  "%s"' % (topdir1, topdir2))
    return tuple(topdirs2)

# Yield all the files in or below "topdir". Exclude UNIX hidden files,
# hidden directories and their contents, and all files with extensions in
# the list "excludes". Include only files with extensions in the list
# "includes". Case is ignored for the includes and excludes. If an extension
# is in both lists, it is excluded. The files are walked in alphabetical
# order. Directories in "excluded_dirs" are not walked.
def Walk(topdirs, includes=None, excludes=None, excluded_dirs=None):
    topdirs = prepare_topdirs(topdirs)
    topdirs = list(topdirs)
    topdirs.sort()
    if includes != None:
        includes2 = set()
        for include in includes:
            includes2.add(include.lower())
        includes = includes2
    if excludes is None:
        excludes = tuple()
    excludes2 = set()
    for exclude in excludes:
        excludes2.add(exclude.lower())
    excludes = excludes2
    excluded_dirs = prepare_excludes(excluded_dirs)
    for topdir in topdirs:
        for thisdir, dirs, files in os.walk(topdir):
            dirs.sort()
            bad = list()
            for i, name in enumerate(dirs):
                fullname = os.path.join(thisdir, name)
                if name[0] == '.' or fullname in excluded_dirs:
                    # UNIX hidden directory or excluded dir.
                    bad.append(i)
            bad.reverse()
            for i in bad:
                del dirs[i]
    
            files.sort()
            for name in files:
                fullname = os.path.abspath(os.path.join(thisdir, name))
                basename = os.path.basename(fullname)
                junk, ext = os.path.splitext(fullname)
                ext = ext.lower()
                extOK = False
                if includes == None or ext in includes:
                    extOK = True
                if name[-1] != '~' and basename[0] != '.' \
                        and not ext in excludes and extOK:
                    yield fullname

# Yield all image files in or below "directory". Exclude UNIX hidden files,
# hidden directories and their contents, and all files with extensions in
# the list "excludes". Include only files with extensions in the list
# "includes". "includes" defaults to the standard image extensions. Case is
# ignored for the includes and excludes. If an extension is in both lists,
# it is excluded. Directories in "excluded_dirs" are not walked.
def WalkImages(topdirs, includes=None, excludes=None, excluded_dirs=None):
    # These are the extensions that can be read by PIL.
    standard_includes = ('.bmp', '.dib', '.cur', '.dcx', '.eps', '.ps',
         '.fli', '.flc', '.fpx', '.gbr', '.gd', '.gif', '.ico', '.im',
         '.jpg', '.jpe', '.jpeg', '.mic', '.msp', '.pcd', '.pcx', '.pdf',
         '.png', '.pbm', '.pgm', '.ppm', '.psd', '.bw', '.rgb', '.cmyk',
         '.ras', '.tga', '.tif', '.tiff', '.xbm', '.xpm')
    if includes is None:
        includes = standard_includes
    else:
        includes = set(standard_includes) | set(includes)
    for filename in Walk(topdirs, includes, excludes, excluded_dirs):
        yield filename

# Yield all the directories at or below the "topdirs". Exclude UNIX hidden
# directories and their contents. The directories are walked in alphabetical
# order.
def WalkDirs(topdirs):
    topdirs = prepare_topdirs(topdirs)
    topdirs = list(topdirs)
    topdirs.sort()
    for topdir in topdirs:
        for thisdir, dirs, files in os.walk(topdir):
            dirs.sort()
            bad = list()
            for i, name in enumerate(dirs):
                fullname = os.path.join(thisdir, name)
                if name[0] == '.':
                    # UNIX hidden directory.
                    bad.append(i)
            bad.reverse()
            for i in bad:
                del dirs[i]
            yield thisdir
