
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import time, sys
import haarc, mrquery, utilities
from constants import HAARSIZE, RCQUOTA, SAVETIME, PRINTAT
import Image, numarray
from numarray.numerictypes import *
from LN import LN, MLLN

class Haar(object):
    """Haar transforms image thumbnails"""
    def __init__(self, topdatadir, basename, topdirs=None,
                 excluded_dirs=None):
        self.started = False
        self.mrq = mrquery.FetchOrCreate_MRQuery(topdatadir, basename, 'as_is')
        self.mrq.ddl_sortlist = list()
        print 'Finished creating or reading MRQuery. ' \
                        'There are %i images in MRQuery.' % \
                 len(self.mrq.hash2ImData)

        rgb2yiq = ((.299,  .587,  .114),
                   (.596, -.275, -.321),
                   (.212, -.523,  .311))
        self.rgb2yiq = numarray.array(rgb2yiq, Float64)

        # Combine row and column into one number.
        ind = numarray.indices( (HAARSIZE, HAARSIZE), Int32)
        self.rcarr = HAARSIZE * ind[0] + ind[1]

        haarc.start(HAARSIZE)
        self.started = True
        self.T1 = 0.0
        self.T2 = 0.0
        self.T3 = 0.0

    def __del__(self):
        """del a Haar object"""
        if hasattr(self, 'started') and self.started:
            haarc.finish()

    def prepim(self, name, im):
        """Prepare an image for the Haar transform"""
        if im.mode not in ('L', 'P', 'RGB'):
            raise TypeError('image mode must be "L", "P" or "RGB"')

        # If a b&w image is 'RGB' or 'P', the "I" and "Q" planes will not
        # be very useful. Need to figure out how to treat these images as b&w.
        # Making a b&w jpeg from every gif did not help.
        if im.mode == 'P':
            thumb = im.convert('RGB')
        elif len(im.mode) > 1 and im.mode != 'RGB':
            thumb = im.convert('RGB')
        else:
            thumb = im.copy()
        thumb.thumbnail( (HAARSIZE, HAARSIZE), Image.ANTIALIAS)
        arr = utilities.PIL_To_NA_Convert(thumb, Float64)
        if len(arr.shape) == 2:
            arr.shape = (1,) + arr.shape

        # Put the thumbnail inside a 128 X 128 array.
        square = numarray.zeros((arr.shape[0], HAARSIZE, HAARSIZE), Float64)
        d = arr.shape[1]
        rlo = (HAARSIZE - d)//2
        rhi = rlo + d
        d = arr.shape[2]
        clo = (HAARSIZE - d)//2
        chi = clo + d
        square[:, rlo:rhi, clo:chi] = arr

        # "mask" is one outside the thumbnail.
        mask = numarray.ones((arr.shape[0], HAARSIZE, HAARSIZE), Int32)
        mask[:, rlo:rhi, clo:chi] = numarray.zeros(arr.shape, Int32)

        # Normalize the coefficients.
        square = square / 256.0
        if square.shape[0] == 1:
            # For a b&w image, the I and Q planes are zeros.
            return square, mask, rlo, rhi, clo, chi
        else:
            square = numarray.transpose(square, (1,2,0))
            square =  numarray.innerproduct(self.rgb2yiq, square)
            return square, mask, rlo, rhi, clo, chi

    def haar_channel(self, arr, mask):
        """Do a Haar transform on one images plane"""
        # "haar_channel" should be moved to C or C++.
        t0 = time.time()
        arr, mask = haarc.haarc(arr, mask)
        t1 = time.time()
        # I don't use the average intensity.
        arr[0, 0] = 0.0
        arr = numarray.where(mask, 0.0, arr)
        # Encode the sign of arr in rcarr.
        rcarr = numarray.where(arr > 0.0, self.rcarr, -self.rcarr)
        # The minus sign reverses the sort.
        arr = -numarray.absolute(arr)
        rcarr = numarray.ravel(rcarr)
        arr = numarray.ravel(arr)
        ind = numarray.argsort(arr)
        ind = ind[:RCQUOTA]
        rcarr = rcarr[ind]
        t2 = time.time()
        rcarr = numarray.sort(rcarr)
        t3 = time.time()
        # Roughly T1 : T2 : T3 = 9 : 6 : .1
        # Actually 45.816 30.847 0.426 for 2650 images.
        # Might want to put numarray stuff in C.
        self.T1 += t1 - t0
        self.T2 += t2 - t1
        self.T3 += t3 - t2
        return rcarr

    def Darr_join(self, Darr_list):
        """Concatenate Darr_list to get self.mrq.Darrs"""
        Darrs = numarray.concatenate(Darr_list)
        Darrs = numarray.transpose(Darrs, (1,0,2))
        self.mrq.Darrs = \
               numarray.concatenate( (self.mrq.Darrs, Darrs), 1)

    def bwbits_join(self, bwbit_list):
        """Concatenate bwbit_list to get self.mrq.bwbits"""
        self.mrq.bwbits = numarray.concatenate(
                          [self.mrq.bwbits] + bwbit_list )

    def all_channels(self, ID):
        """Haar tranform the image corresponding to an ImData object"""
        pathname = ID[0]
        im = Image.open(pathname)
        arr, mask, rlo, rhi, clo, chi = self.prepim(pathname, im)
        trip = list()
        for layer in range(arr.shape[0]):
            rcarr = self.haar_channel(arr[layer], mask[layer])
            trip.append(rcarr)
        bwbit = (arr.shape[0] == 1)
        if bwbit:
            Darr = numarray.zeros((3, 40), Int32)
            Darr[0:1, :] = numarray.array(trip, type=Int32)
        else:
            Darr = numarray.array(trip, type=Int32)
        Darr.shape = (1, 3, 40)
        return Darr, bwbit

    def mainloop(self):
        """Haar tranforms a set if images"""
        template1 = 'Haar: new ims: %i, total ims: %i, %3.2f secs/im, %3.2f secs total.\n  At %s'
        template2 = 'Haar: %i images added, %i total, %3.2f secs/im, %3.2f secs total.'
        template3 = 'Haar: No images done, %3.2f secs total.'
        template4 = 'Haar completed: there are %i images.'

        savetime = t0 = t1 = time.time()
        Darr_list = list()
        bwbit_list = list()
        count = 0
        zilch = numarray.zeros( (1, 3, 40), Int32)
        firsthaar = self.mrq.haar_count
        # Loop over the new imnums that will be needed.
        LN('Haar will be run on approximately %s images.' %
                       (self.mrq.imnum_count - firsthaar))
        for imnum in range(firsthaar, self.mrq.imnum_count):
            if imnum not in self.mrq.imnum2hash:
                # I will use imnum to index self.mrq.Darrs.
                LN('imnum', imnum, 'not used.')
                Darr_list.append(zilch)
                bwbit_list.append(0)
                self.mrq.haar_count += 1
                continue
            hashcode = self.mrq.imnum2hash[imnum]
            ID = self.mrq.hash2ImData[hashcode]
            # assert not ID.deleted
            # Haar transform the image.
            Darr, bwbit = self.all_channels(ID)
            Darr_list.append(Darr)
            bwbit_list.append(bwbit)
            self.mrq.haar_count += 1
            count += 1
            if time.time() - savetime >= SAVETIME:
                tt0 = time.time()
                self.Darr_join(Darr_list)
                self.bwbits_join(bwbit_list)
                self.mrq.Store()
                print 'Store took %4.2f secs.' % (time.time() - tt0)
                Darr_list = list()
                bwbit_list = list()
                savetime = time.time()
            if count % PRINTAT == 0:
                t4 = time.time()
                text = template1 % (count, imnum+1, (t4-t1)/PRINTAT, t4-t0,
                      ID[0].as_str())
                t1 = time.time()
                print >> sys.stderr, text
                text2 = 'Haar timings haarc: %5.3f, mumarray: %5.3f, sort: %5.3f' % \
                           (self.T1, self.T2, self.T3)
                print >> sys.stderr, text2

        tt = time.time()
        if len(Darr_list) > 0:
            self.Darr_join(Darr_list)
            self.bwbits_join(bwbit_list)
            self.mrq.clean()
            self.mrq.Store()
        if count == 0:
            text = template3 % (tt-t0)
            print >> sys.stderr, text
        else:
            text = template2 % (count, imnum+1, (tt-t0)/count, tt-t0)
            print >> sys.stderr, text
        text = template4 % len(self.mrq.hash2ImData)
        print >> sys.stderr, text
