#! /usr/bin/env python

"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os, sys, math, numarray, IM
from numarray.numerictypes import *
from LN import LN, MLLN

RSQRT2 = 0.70710678118654752440
HAARSIZE = 128
GEN = 7

"""2-d Haar transform.

"arr" is a Float64 numarray of shape (haarsize, haarsize).
"mask" is an Int32 numarray of the same shape. The pixels to be masked
out have value "1", the others, "0".

A pixel is "bad" if it is masked out, otherwise "good". When summing a group
of pixels, treat each bad pixel as if it equalled the average of the good
pixels. Let m and g be the number of pixels and the number of good pixels,
respectively. Let S be the sum over the good pixels. Result = m*S/g.

My C version of this puts all the blocksums into a larger array (256x256).
Then the + - stuff is done. Suppose HAARSIZE = 128. Orig is "thumb". "sumarr"
has shape (256, 256). "thumb" is copied to sumarr[128:, 128:].
sumarr[128:, 128:] is blocksummed by (1,2) and put into sumarr[128:, 64:128].
This is continued to the left, filling up the botton half of sumarr. Then this
blocksumming process is done vertically, filling up the rest of sumarr.
Row 0 and col 0 of sumarr are not used. The square roots are done by each
block sum. Similarly for the mask.
"""
def counts_basic(h):
    # Coefficient 0 is the sum of h numbers. Coefficient 1 is X - Y
    # where X and Y are the sums of h/2 numbers. Coefficients 2 and 3
    # have the form X - Y where X and Y are the sums of h/4 terms. ...
    # For h = 16, counts returns
    #       [16, 8, 4, 4, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1]
    cts = h * [0]
    cts[0] = h
    i = 1
    n = 1
    while h > 1:
        h /= 2
        lo = 0
        for j in range(n):
            cts[i] = h
            lo += 2*h
            i += 1
        n *= 2
    return cts

def haar(arr, mask, haarsize=HAARSIZE, gen=GEN):
    """2-d Haar transform.

    This is a Python versiojn of the C code used to develop the algorithm.

    "arr" is a Float64 numarray of shape (haarsize, haarsize).
    "mask" is an Int32 numarray of the same shape. The pixels to be masked
    out have value "1", the others, "0".

    A pixel is "bad" if it is masked out, otherwise "good". When summing a
    group of pixels, treat each bad pixel as if it equalled the average of
    the good pixels. Let m and g be the number of pixels and the number of
    good pixels, respectively. Let S be the sum over the good pixels.
    Result = m*S/g.
    """
    mask = numarray.asarray(mask, Int32)
    arr = numarray.asarray(arr, Float64)
    arr =  numarray.where(mask, 0.0, arr)
    # I prefer to count good pixels rather than bad ones. So negate the mask.
    mask = numarray.logical_not(mask)
    # Row 0 and column 0 or sumarr and masksum are not used. This makes the
    # indexing easier.
    totals = counts_basic(2*haarsize)
    totals[0] = 0
    totals = numarray.outerproduct(totals, totals)
    sumarr = numarray.zeros( (2*haarsize, 2*haarsize), Float64) 
    masksum = numarray.zeros( (2*haarsize, 2*haarsize), Int32)
    sumarr[haarsize : 2*haarsize, haarsize : 2*haarsize] = arr
    masksum[haarsize : 2*haarsize, haarsize : 2*haarsize] = mask

    # Fill the lower half of sumarr and masksum.
    rowslice = slice(haarsize, 2*haarsize)
    for n in reversed(range(0, gen)):
        nslice = slice(2**n, 2**(n+1))
        prevslice_even = slice(2**(n+1), 2**(n+2), 2)
        prevslice_odd = slice(2**(n+1)+1, 2**(n+2), 2)
        # Whenever the Haar transform is calculated recursively, an
        # L_2 normalization factor of 1/sqrt(2) is put in at each stage.
        sumarr[rowslice, nslice] = \
           (sumarr[rowslice, prevslice_even] +
            sumarr[rowslice, prevslice_odd]) * RSQRT2
        masksum[rowslice, nslice] = \
            masksum[rowslice, prevslice_even] + \
            masksum[rowslice, prevslice_odd]

    # Fill the upper half of sumarr and masksum.
    for n in reversed(range(0, gen)):
        for r in range(2**n, 2**(n+1)):
            sumarr[r, :] = (sumarr[2*r, :] + sumarr[2*r+1, :]) * RSQRT2
            masksum[r, :] = masksum[2*r, :] + masksum[2*r+1, :]

    # Prevents divides by zero.
    masksum2 = numarray.maximum(masksum, 1)
    fish = totals.astype(Float64) / masksum2.astype(Float64)
    fish = numarray.where(masksum > 0, fish, 0.0)

    arrout = numarray.zeros((haarsize, haarsize), Float64)
    maskout = numarray.zeros((haarsize, haarsize), Int32)

    # Fill arrout except for the first row and column. 2 by 2 blocks of pixels
    # in sumarr are used.
    #    arrout[r, c] = (sumarr[2*r : 2*c] - sumarr[2*r+1 : 2*c] -
    #              sumarr[2*r : 2*c+1] + sumarr[2*r+1 : 2*c+1])/sqrt(4)
    evens = slice(2, 2*haarsize, 2)
    odds = slice(3, 2*haarsize, 2)
    arrout[1:, 1:] = (
            sumarr[evens, evens] * fish[evens, evens] / 2.0
          - sumarr[odds, evens] * fish[odds, evens] / 2.0
          - sumarr[evens, odds] * fish[evens, odds] / 2.0
          + sumarr[odds, odds] * fish[odds, odds] / 2.0)

    # If masksum is zero at any of the four pixels in a 2 by 2 block,
    # maskout will be zero.
    masksum = (masksum > 0)
    maskout[1:, 1:] =  masksum[evens, evens] * masksum[odds, evens] * \
          masksum[evens, odds] * masksum[odds, odds]

    arrout[0, 0] = sumarr[1, 1] * fish[1, 1]
    maskout[0, 0] = min(masksum[1, 1], 1)

    arrout[1:, 0:1] =   sumarr[evens, 1:2] * fish[evens, 1:2] * RSQRT2 \
                      - sumarr[odds, 1:2] * fish[odds, 1:2] * RSQRT2
    maskout[1:, 0:1] = numarray.logical_and(
          masksum[evens, 1:2], masksum[odds, 1:2])

    arrout[0:1, 1:] =   sumarr[1:2, evens] * fish[1:2, evens] * RSQRT2 \
                      - sumarr[1:2, odds] * fish[1:2, odds] * RSQRT2
    maskout[0:1, 1:] = numarray.logical_and(
          masksum[1:2, evens], masksum[1:2, odds])

    arrout = numarray.where(maskout, arrout, 0.0)
    maskout = numarray.logical_not(maskout)
    return arrout, maskout

def test2():
#    seq = ((1.0, 3.0), (0.0, 2.0))
    seq = ((1, 3, 2, -1), (0, -2, -3, 3), (5, 4, 0, -2), (-5, -4, 2, 4))
    arr = numarray.array(seq, Float64)
    mask = numarray.zeros(arr.shape, Int32)
#    mask[0,0] = 1
    outarr, maskarr = haar(arr, mask, 4, 2)
    print maskarr
    print outarr

if __name__ == '__main__':
    numarray.Error.pushMode(all="raise")
    test2()
