
/* Copyright 2006 Edward C. Jones

   This file is part of mrquery.

   mrquery is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.

   mrquery is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with mrquery; if not, write to the Free Software Foundation, Inc., 51
   Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "haarc.h"

static int haarsize = -1;
static int generation = -1;
static Float64 **sumarr = NULL;
static Int32 **good = NULL;
static Float64 **fish = NULL;
static Float64 *counts = NULL;
static Float64 **maxgood_wted = NULL;
static int available[4] = {1,1,1,1};
static int was_started = 0;
// static int is_first_start = 1;

/*
  2-d Haar transform.

  "arr" is a Float64 numarray of shape (haarsize, haarsize).
  "mask" is an Int32 numarray of the same shape. The pixels to be masked
  out have value "1", the others, "0".

  A pixel is "bad" if it is masked out, otherwise "good". When summing a
  group of pixels, treat each bad pixel as if it equalled the average of the
  good pixels. Let m and g be the number of pixels and the number of good
  pixels, respectively. Let S be the sum over the good pixels. Result =
  m*S/g.

  This C version of the above puts all the blocksums into a larger array
  (256x256). Then the + - stuff is done. Suppose HAARSIZE = 128. Orig is
  "thumb". "sumarr" has shape (256, 256). "thumb" is copied to
  sumarr[128:, 128:]. sumarr[128:, 128:] is blocksummed by (1,2) and put
  into sumarr[128:, 64:128]. This is continued to the left, filling up the
  botton half of sumarr. Then this blocksumming process is done vertically,
  filling up the rest of sumarr. Row 0 and col 0 of sumarr are not used. 
  The square roots are done by each block sum. Similarly for the mask.
*/

/*=== Range checks =========================================*/

/*
static PyObject*
MyArray_Check1(PyArrayObject *arr, int r)
{
    if (r < 0 || r >= arr->dimensions[0]) {
        PyErr_SetString(PyExc_IndexError, "subscript out of range");
        return NULL;
    }
    Py_RETURN_NONE;
}

static PyObject*
MyArray_Check2(PyArrayObject *arr, int r, int c)
{
    if (r < 0 || r >= arr->dimensions[0] || c < 0 || c >= arr->dimensions[1]) {
        PyErr_SetString(PyExc_IndexError, "subscript out of range");
        return NULL;
    }
    Py_RETURN_NONE;
}

static PyObject*
rangecheck1(int r, int minr, int maxr)
{
    if (r < minr || r >= maxr) {
        PyErr_SetString(PyExc_IndexError, "subscript out of range");
        return NULL;
    }
    Py_RETURN_NONE;
}

static PyObject*
rangecheck2(int r, int minr, int maxr, int c, int minc, int maxc)
{
    if (r < minr || r >= maxr || c < minc || c >= maxc) {
        PyErr_SetString(PyExc_IndexError, "subscript out of range");
        return NULL;
    }
    Py_RETURN_NONE;
}
*/

/*=== Utilities ============================================*/

static PyObject*
free_globals(void)
{
    PyMem_Free(counts); counts = NULL;
    if (available[MAXGOOD_WTED] == 0) {
        free_arrayD(maxgood_wted);
        maxgood_wted = NULL;
        available[MAXGOOD_WTED] = 1;
    }
    if (available[FISH] == 0) {
        free_arrayD(fish);
        fish = NULL;
        available[FISH] = 1;
    }
    if (available[GOOD] == 0) {
        free_arrayI(good);
        good = NULL;
        available[GOOD] = 1;
    }
    if (available[SUMARR] == 0) {
        free_arrayD(sumarr);
        sumarr = NULL;
        available[SUMARR] = 1;
    }
    haarsize = -1;
    generation = -1;
    return NULL;
}

double**
arrayD(int rows, int cols)
{
    int r;
    double** arr = NULL;

    arr = PyMem_Malloc(rows * sizeof(double));
    if (arr == NULL) {
        PyErr_NoMemory();
        return NULL;
    }
    arr[0] = PyMem_Malloc(rows * cols * sizeof(double));
    if (arr[0] == NULL) {
        PyMem_Free(arr);
        arr = NULL;
        PyErr_NoMemory();
        return NULL;
    }
    zero_arrayD(arr, rows, cols);
    for (r=1; r<rows; r++) {
        arr[r] = arr[r-1] + cols;
    }
    return arr;
}

void
zero_arrayD(double** arr, int rows, int cols)
{
    memset(arr[0], 0, rows * cols * sizeof(double));
}

void
free_arrayD(double** arr)
{
    if (arr != NULL) {
        if (arr[0]) PyMem_Free(arr[0]);
        arr[0] = NULL;
        PyMem_Free(arr);
    }
}

int**
arrayI(int rows, int cols)
{
    int r;
    int** arr = NULL;

    arr = PyMem_Malloc(rows * sizeof(int));
    if (arr == NULL) {
        PyErr_NoMemory();
        return NULL;
    }
    arr[0] = PyMem_Malloc(rows * cols * sizeof(int));
    zero_arrayI(arr, rows, cols);
    if (arr[0] == NULL) {
        PyMem_Free(arr);
        arr = NULL;
        PyErr_NoMemory();
        return NULL;
    }
    for (r=1; r<rows; r++)
        arr[r] = arr[r-1] + cols;
    return arr;
}

void
zero_arrayI(int** arr, int rows, int cols)
{
    memset(arr[0], 0, rows * cols * sizeof(int));
}

/* As part of free_arrayI, I want to set arr = NULL. But I can only do this
   if I pass free_arrayI the _address_ of arr. */
void
free_arrayI(int** arr)
{
    if (arr) {
        if (arr[0]) PyMem_Free(arr[0]);
        arr[0] = NULL;
        PyMem_Free(arr);
    }
}

/*=== Print functions ===================================*/

/*
static PyObject*
MyArray_Print(PyArrayObject *arr, char* name, int lineno)
{
    int r, c;

    if (arr->nd != 2) {
        PyErr_SetString(PyExc_TypeError, "array must be 2-d");
        return NULL;
    }
    char t[10];
    t[0] = 0;
    if (lineno >= 0)
        sprintf(t, "L%i ", lineno);
    printf("*%s%s:\n", t, name);

    for (r=0; r < arr->dimensions[0]; r++) {
        for (c=0; c < arr->dimensions[1]; c++) {
            if (arr->descr->type_num == tFloat64)
                printf("%f ", PIX2(arr, r, c, Float64));
            else if (arr->descr->type_num == tInt32)
                printf("%i ", PIX2(arr, r, c, Int32));
            else {
                PyErr_SetString(PyExc_TypeError,
                  "array must have type Int32 or Float64");
                return NULL;
            }
        }
        printf("\n");
    }
    Py_RETURN_NONE;
}        

static void
printarr_Float64(Float64 **arr, int rhi, int chi, char* name)
{
    int r, c;

    printf("%s:\n", name);

    for (r=0; r < rhi; r++) {
        for (c=0; c < chi; c++)
            printf("%g ", arr[r][c]);
        printf("\n");
    }
}        

static void
printarr_Int32(Int32 **arr, int rhi, int chi, char* name)
{
    int r, c;

    printf("*%s:\n", name);

    for (r=0; r < rhi; r++) {
        for (c=0; c < chi; c++)
            printf("%i ", arr[r][c]);
        printf("\n");
    }
}        

static void
printvec_Float64(Float64 *vec, int hi, char* name)
{
    int r;

    printf("*%s:\n", name);

    for (r=0; r < hi; r++)
        printf("%g ", vec[r]);
    printf("\n");
}


static void
printvec_Int32(Int32 *vec, int hi, char* name)
{
    int r;

    printf("*%s:\n", name);

    for (r=0; r < hi; r++)
        printf("%i ", vec[r]);
    printf("\n");
}
*/

/*=== Haar transform =======================================*/

static void
fill_lower_Float64(Float64 **arr)
{
    int n, r, c;

    for (n = generation-1; n >= 0; n--)
        for (r = haarsize; r < 2*haarsize; r++)
            for (c = POW2(n); c < POW2(n+1); c++)
                arr[r][c] = arr[r][2*c] + arr[r][2*c+1];
}

static void
fill_lower_Int32(Int32 **arr)
{
    int n, r, c;

    for (n = generation-1; n >= 0; n--)
        for (r = haarsize; r < 2*haarsize; r++)
            for (c = POW2(n); c < POW2(n+1); c++)
                arr[r][c] = arr[r][2*c] + arr[r][2*c+1];
}

static void
fill_upper_Float64(Float64 **arr)
{
    int n, r, c;

    for (n=generation-1; n>=0; n--)
        for (c=1; c<2*haarsize; c++)
            for (r=POW2(n); r<POW2(n+1); r++)
                arr[r][c] = arr[2*r][c] + arr[2*r+1][c];
}

static void
fill_upper_Int32(Int32 **arr)
{
    int n, r, c;

    for (n=generation-1; n>=0; n--)
        for (c=1; c<2*haarsize; c++)
            for (r=POW2(n); r<POW2(n+1); r++)
                arr[r][c] = arr[2*r][c] + arr[2*r+1][c];
}

static PyObject*
Chaar(PyArrayObject* arr, PyArrayObject* mask, PyArrayObject* arrout,
         PyArrayObject* maskout)
{
    int r, c, pixels;

    pixels = 2*haarsize * 2*haarsize;
    zero_arrayD(sumarr, 2*haarsize, 2*haarsize);
    zero_arrayI(good, 2*haarsize, 2*haarsize);
    zero_arrayD(fish, 2*haarsize, 2*haarsize);

    /* Copy "arr" into "sumarr". */
    for (r=0; r<haarsize; r++)
        for (c=0; c<haarsize; c++)
            sumarr[r+haarsize][c+haarsize] = PIX2(arr, r, c, Float64);

    /* Copy "mask" into "good". In the input mask, the pixels to be masked are
       zeros, the others are ones. We want the reverse. */
    for (r=0; r<haarsize; r++) {
        for (c=0; c<haarsize; c++) {
            good[r + haarsize][c + haarsize] = 
                 !PIX2(mask, r, c, Int32);
        }
    } 

    fill_lower_Int32(good);
    fill_upper_Int32(good);
    fill_lower_Float64(sumarr);
    fill_upper_Float64(sumarr);

    /* Find the "fish" factors. A pixel is "bad" if it is masked out,
       otherwise "good". When summing a group of pixels, treat each bad pixel
       as if it equalled the average of the good pixels. Let m and g be the
       number of pixels and the number of good pixels, respectively. Let S be
       the sum over the good pixels. Result = m*S/g.
    */
    for (r=1; r<2*haarsize; r++) {
        for (c=1; c<2*haarsize; c++) {
            if (good[r][c] == 0)
                fish[r][c] = 0.0;
            else
                fish[r][c] = counts[r/2] * counts[c/2] / (Float64) good[r][c];
        }
    }

    /* If "good" is zero at any of the four pixels in a 2 by 2 block,
        "maskout" will be zero. */
    int itemp;
    /* r > 1, c > 1. */
    for (r=1; r<haarsize; r++) {
        for (c=1; c<haarsize; c++) {
            /* If good is zero at any of the four pixels in a 2 by 2 block,
               maskout will be zero. */
            itemp = good[2*r][2*c] && good[2*r+1][2*c] &&
                    good[2*r][2*c+1] && good[2*r+1][2*c+1];
            /* We need to do a logical_not on "maskout". Might as well do it
               here. */
            PIX2(maskout, r, c, Int32) = !itemp;

            if (itemp == 0)
                PIX2(arrout, r, c, Float64) = 0.0;
            else {
                /* Fill "arrout" except for the first row and column. 2 by 2
                   blocks of pixels in sumarr are used. Note that 1/2 =
                   (1/sqrt(2))**2. */
                PIX2(arrout, r, c, Float64) = 
                    + sumarr[2*r][2*c] * fish[2*r][2*c]
                    - sumarr[2*r+1][2*c] * fish[2*r+1][2*c]
                    - sumarr[2*r][2*c+1] * fish[2*r][2*c+1]
                    + sumarr[2*r+1][2*c+1] * fish[2*r+1][2*c+1];
            }
        }
    }

    /* c == 0 edge. */
    for (r=1; r<haarsize; r++) {
        itemp = good[2*r][1] && good[2*r+1][1];
        PIX2(maskout, r, 0, Int32) = !itemp;
        if (itemp == 0)
            PIX2(arrout, r, 0, Float64) = 0.0;
        else
            PIX2(arrout, r, 0, Float64) =
                 + sumarr[2*r][1] * fish[2*r][1]
                 - sumarr[2*r+1][1] * fish[2*r+1][1];
    }

    /* r == 0 edge. */
    for (c=1; c<haarsize; c++) {
        itemp = good[1][2*c] && good[1][2*c+1];
        PIX2(maskout, 0, c, Int32) = !itemp;
        if (itemp == 0)
            PIX2(arrout, 0, c, Float64) = 0.0;
        else
            PIX2(arrout, 0, c, Float64) =
                 + sumarr[1][2*c] * fish[1][2*c]
                 - sumarr[1][2*c+1] * fish[1][2*c+1];
    }

    /* (0, 0) corner. */
    itemp = good[1][1] ? 1 : 0;
    PIX2(maskout, 0, 0, Int32) = !itemp;
    if (itemp == 0)
        PIX2(arrout, 0, 0, Float64) = 0.0;
    else
        PIX2(arrout, 0, 0, Float64) = sumarr[1][1] * fish[1][1];

    /* Do the 1/sqrt(2) normalizations. */
    for (r=0; r<haarsize; r++)
        for (c=0; c<haarsize; c++)
            PIX2(arrout, r, c, Float64) /= maxgood_wted[r][c];
    Py_RETURN_NONE;
}

static PyObject*
haarc(PyObject* obj, PyObject *args)
{
    PyObject* oarr = NULL;
    PyObject* omask = NULL;
    PyArrayObject* arr = NULL;
    PyArrayObject* mask = NULL;
    PyArrayObject* arrout = NULL;
    PyArrayObject* maskout = NULL;
    PyObject* pyo = NULL;
    PyObject* pair = NULL;

    if (!was_started) {
        fflush(stdout);
        fflush(stdout);
        PyErr_SetString(PyExc_RuntimeError, "\"start\" was never called");
        return NULL;
    }

    if (!PyArg_ParseTuple(args, "OO:haar", &oarr, &omask))
        goto fail;
    arr = NA_InputArray(oarr, tFloat64, NUM_C_ARRAY);
    if (arr == NULL) goto fail;
    mask = NA_InputArray(omask, tInt32, NUM_C_ARRAY);
    if (mask == NULL) goto fail;
    arrout = NA_NewArray(NULL, tFloat64, 2, haarsize, haarsize);
    if (arrout == NULL) goto fail;
    maskout = NA_NewArray(NULL, tInt32, 2, haarsize, haarsize);
    if (maskout == NULL) goto fail;
    pyo = Chaar(arr, mask, arrout, maskout);
    if (pyo == NULL) goto fail;
    pair = PyTuple_Pack(2, (PyObject*) arrout, (PyObject*) maskout);
    if (pair == NULL) goto fail;

    // PyTuple_Pack INCREFs each item in the tuple.
    Py_XDECREF(maskout);
    Py_XDECREF(arrout);
    // Required by numarray.
    Py_XDECREF(mask);
    Py_XDECREF(arr);

    return pair;

fail:
    Py_XDECREF(pair);
    Py_XDECREF(maskout);
    Py_XDECREF(arrout);
    Py_XDECREF(mask);
    Py_XDECREF(arr);
    return free_globals();
}

/*=== Haar Channel ==============================================*/

/*
Incomplete

static PyObject*
Chaar_channel(PyArrayObject* arr, PyArrayObject* mask, PyArrayObject* rcarr)
{
    int ok;
    PyArrayObject* arr2;
    PyArrayObject* mask2;
    PyArrayObject* values;

    arr2 = NA_NewArray(NULL, tFloat64, 2, haarsize, haarsize);
    if (arrout == NULL) goto fail;
    mask2 = NA_NewArray(NULL, tInt32, 2, haarsize, haarsize);
    if (maskout == NULL) goto fail;
    values = NA_NewArray(NULL, tFloat64, 2, haarsize*haarsize, 4);
    if (values == NULL) goto fail;

    if (!Chaar(arr, mask, arr2, mask2))
        goto fail;
    // I don't use the avarage intensity.
    PIX2(arr2, 0, 0, Float64) = 0.0;
*/

/*=== Start and finish ==========================================*/

static void
counts_basic(void)
{
    int i, n, h, j, r, c;

    /* Coefficient 0 is the sum of h numbers. Coefficient 1 is X - Y
       where X and Y are the sums of h/2 numbers. Coefficients 2 and 3
       have the form X - Y where X and Y are the sums of h/4 terms. ...
       For h = 16, counts returns
            [16, 8, 4, 4, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1]
    */
    counts[0] = haarsize;
    i = 1;
    n = 1;
    h = haarsize;
    while (h > 1) {
        h >>= 1;
        for (j=0; j<n; j++) {
            counts[i++] = (Float64) h;
        }
        n <<= 1;
    }

    /* See ctssqrt in thetruth.py. */
    for (r=1; r<haarsize; r++) {
        for (c=1; c<haarsize; c++) {
            maxgood_wted[r][c] = 2.0 * sqrt(counts[r] * counts[c]);
        }
    }
    for (r=1; r<haarsize; r++) {
        maxgood_wted[r][0] = M_SQRT2 * sqrt(counts[r] * counts[0]);
    }
    for (c=1; c<haarsize; c++) {
        maxgood_wted[0][c] = M_SQRT2 * sqrt(counts[0] * counts[c]);
    }
    maxgood_wted[0][0] = counts[0];
}

/* Run this soon after haar.so is imported. No need to run it for each
   image. */
static PyObject*
start(PyObject* obj, PyObject *args)
{
//    if (is_first_start) {
//        is_first_start = 0;
//        mtrace();
//        int* leak = (int*) malloc(sizeof(int));
//    }
    /* It is permitted to call start more than once. It is nicer to call
       finish before calling start again. Clean up globals. */
    free_globals();

    if (!PyArg_ParseTuple(args, "i:start", &haarsize))
        goto fail;
    /* haarsize and generation are globals. */

    int prod = 1;
    for (generation=1; generation<12; generation++) {
        prod *= 2;
        if (prod == haarsize)
            break;
    }
    if (generation >= 10)
        PyErr_SetString(PyExc_ValueError, "haarsize must be 2, 4, 8, ..., 1024");

    sumarr = arrayD(2*haarsize, 2*haarsize);
    if (sumarr == NULL) goto fail;
    available[SUMARR] = 0;
    good = arrayI(2*haarsize, 2*haarsize);
    if (good == NULL) goto fail;
    available[GOOD] = 0;
    fish = arrayD(2*haarsize, 2*haarsize);
    if (fish == NULL) goto fail;
    available[FISH] = 0;
    maxgood_wted = arrayD(haarsize, haarsize);
    available[MAXGOOD_WTED] = 0;
    counts = PyMem_Malloc(haarsize * sizeof(Float64));
    if (!counts) goto fail;

    memset(counts, 0, haarsize * sizeof(Float64));

    counts_basic();
    was_started = 1;
    Py_RETURN_NONE;

fail:
    return free_globals();
}

/* Run finish after your Python program has processed all the images. */
PyObject*
finish(PyObject* obj, PyObject *args)
{
    if (!was_started) {
        fflush(stderr);
        fflush(stdout);
        PyErr_SetString(PyExc_TypeError, "\"start\" was never called");
        return NULL;
    }
    free_globals();
    Py_RETURN_NONE;
}

/*=== Boilerplate =============================================*/

static PyMethodDef haarc_Methods[] = {
    {"haarc", haarc, METH_VARARGS, "Haar transform"},
    {"start", start, METH_VARARGS, "Initialize globals"},
    {"finish", finish, METH_VARARGS, "DECREF globals"},
    {NULL, NULL, 0, NULL}   /* sentinel */
};

void inithaarc(void)
{
    Py_InitModule("haarc", haarc_Methods);
    import_libnumarray();
}
