#! /usr/bin/env python

"""A Haar transform program that I use to debug other Haar transform programs"""

"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os, sys, math, numarray
from numarray.numerictypes import *
from LN import LN, MLLN

def counts_basic(h):
    # Coefficient 0 is the sum of h numbers. Coefficient 1 is X - Y
    # where X and Y are the sums of h/2 numbers. Coefficients 2 and 3
    # have the form X - Y where X and Y are the sums of h/4 terms. ...
    # For h = 16, counts returns
    #       [16, 8, 4, 4, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1]
    cts = h * [0]
    blks = h * [None]
    cts[0] = h
    blks[0] = ((0,h), (h,h))
    i = 1
    n = 1
    while h > 1:
        h /= 2
        lo = 0
        for j in range(n):
            cts[i] = h
            blks[i] = ((lo, lo+h), (lo+h, lo+2*h))
            lo += 2*h
            i += 1
        n *= 2
    return cts, blks

def ctssqrt(h, counts):
    out = [math.sqrt(counts[0])]
    for i in range(1, h):
        out.append(math.sqrt(2 * counts[i]))
    arr = numarray.outerproduct(out, out).astype(Float64)
    return arr

def blocksum(seq, mask, rpair, cpair, normalize):
    rlo, rhi = rpair
    clo, chi = cpair
    slots = (rhi-rlo)*(chi-clo)
    if slots == 0:
        return 0.0, False
    count = 0
    sum = 0.0
    for i in range(rlo, rhi):
        for j in range(clo, chi):
            if mask[i][j] == 1:
                count += 1
            else:
                sum += seq[i][j]
    if slots == count:
        t = 0
        masked = True
    elif count == 0:
        t = sum
        masked = False
    else:
        if normalize:
            t = float(sum * slots) / float(slots-count)
        else:
            t = sum
        masked = False
    return t, masked

def haar_coeff(seq, mask, r, c, blocks, normalize=True):
    rpair1, rpair2 = blocks[r]
    cpair1, cpair2 = blocks[c]
    a, ma = blocksum(seq, mask, rpair1, cpair1, normalize)
    b, mb = blocksum(seq, mask, rpair1, cpair2, normalize)
    c, mc = blocksum(seq, mask, rpair2, cpair1, normalize)
    d, md = blocksum(seq, mask, rpair2, cpair2, normalize)
    if ma or mb or mc or md:
        return 0, True
    else:
        return a - b - c + d, False

def haar(seq, mask, normalize=True):
    h = len(seq)
    for r in range(h):
        for c in range(h):
            if mask[r][c] == 1:
                seq[r][c] = 0
    counts, blocks = counts_basic(h)
    out1 = []
    mask1 = []
    for r in range(h):
        out2 = []
        mask2 = []
        for c in range(h):
            value, masked = haar_coeff(seq, mask, r, c, blocks, normalize)
            out2.append(value)
            mask2.append(masked)
        out1.append(out2)
        mask1.append(mask2)
    answer = numarray.array(out1, Float64)
    maskout = numarray.array(mask1, UInt8)

    if normalize:
        answer = answer / ctssqrt(h, counts)
    return answer, maskout
